
#include "Material.h"


#include "Body.h"


#include "BodyCollection.h"

#include "Scene.h"

Scene::Scene(BodyCollection* bC, Shader* shader)  : bodyCollection(bC), shader(shader)
{
    shader->setScene(this); //tell the shader to use this scene
}

Scene::~Scene()
{
    delete bodyCollection;
    delete shader;
}

//----------------------generateImage----------------------
// In this function we use the camera to generate rays and
// the shader to color them
//---------------------------------------------------------
void Scene::generateImage() const
{
    //For now we're hacking the ray generation, later we'll use a camera class
    int pxlInRow = 256;
    ImagePPM image(pxlInRow,pxlInRow);
    Vector3d cameraPos(0, -20.0, 0);
    double delta = 10.0/pxlInRow;

    for (int i=0; i<pxlInRow; ++i)
    {
        for (int j=0; j<pxlInRow; ++j)
        {
            //generate ray
            Vector3d pixelPos = Vector3d(-5.0 + j*delta, 0, 5 - i*delta);
            Vector3d dir = pixelPos - cameraPos;
            dir.normalize();
            Ray ray(cameraPos, dir);

            //get color
            Color3d pixelColor = shader->getColor(ray);
            image.setPixel(i, j, truncateToning(pixelColor));
            //image.setPixel(i, j, reinhardMap(pixelColor));
        }
    }
    ofstream out;
    out.open("image.ppm");
    image.writePPM(out);
    out.close();
}

//----------------------reinhardMap------------------------
// This function gets a color with arbitrary float values
// as input and scales them to integers between 0 and 255
// using the Reinhard algorithm:
// $L_1 = 255 \frac{L_0}{1+L_0}$
//---------------------------------------------------------
Color3i Scene::reinhardMap(Color3d& color) const
{
    Color3i output;
    double alpha = 0.5; //arbitrary scaling value.  In the future, this should be related to the average luminence
    for (int i=0; i<3; ++i)
    {
        output[i] = (int) (255 * (color[i] / (color[i] + alpha)));
    }
    return output;
}

//----------------------truncateToning-------------------------
// Simply truncate RGB values over 255
//---------------------------------------------------------
Color3i Scene::truncateToning(Color3d& color) const
{
    Color3i output;
    for (int i=0; i<3; ++i)
    {
        output[i] = (int) clipToInterval(255*color[i], 0, 255);
        //output[i] = (int) (255*color[i]);
    }
    return output;
}

int Scene::loadConfigFile(const char* configFile)
{
    L = lua_open(); //init the lua machine
    luaL_openlibs(L); //open common libs

    //open and load config file
    if (luaL_loadfile(L, configFile) || lua_pcall(L, 0, 0, 0))
    {
        std::cout<<"error: " << lua_tostring(L, -1) << std::endl;
        return 0;
    }
    
    loadLights(); //load the lights from the file
    loadBodies(); //load the bodies
    
    lua_close(L);
    
    return 1;
}

//----------------------loadBodies-------------------------
// Use the Bodies table to load the bodies in the config file
//---------------------------------------------------------
void Scene::loadBodies()
{
    lua_getglobal(L, "bodies"); //push the bodies table onto the stack
    int numBodies = luaL_getn(L, -1); //size of table of bodies

    for (int i=1; i<=numBodies; ++i)
    {
        lua_rawgeti(L, -1, i); //push bodies[i]
        bodyCollection->addBody(loadBody()); //add the current light to the list
        lua_pop(L, 1); //pop bodies[i]
    }
    lua_pop(L, 1); //pop the bodies table
}
//-----------------------loadLights------------------------
// Use the Lights table in the config table to build a list
// of lights
//---------------------------------------------------------
void Scene::loadLights()
{    
    lua_getglobal(L, "lights"); //push the lights table onto the stack
    int numLights = luaL_getn(L, -1); //size of table of lights
    
    for (int i=1; i<=numLights; ++i)
    {
        lua_rawgeti(L, -1, i); //push lights[i]
        addLight(loadLight()); //add the current light to the list
        lua_pop(L, 1); //pop lights[i]
    }

    lua_pop(L, 1); //pop the lights table
}

Body* Scene::loadBody()
{
    double var;

    //first construct the geometry object
    Body* body = new Body();
    Geometry* geo = loadGeometry();
    body->setGeometry(geo);
    body->getMaterial()->setColor(loadColor("color"));

    if (loadDouble("diffuse", var)) //if the diffuse variable exists
    {
        body->getMaterial()->setDiffuse(var);
    }

    if (loadDouble("reflection", var)) //if the diffuse reflection exists
    {
        body->getMaterial()->setReflection(var);
    }

    if (loadDouble("shininess", var)) //if the shininess variable exists
    {
        body->getMaterial()->setShininess(var);
    }

    return body;
}

Geometry* Scene::loadGeometry()
{
    Geometry* geo;

    lua_pushstring(L, "geometry"); //push "geometry" onto the stack
    lua_gettable(L, -2); //get the geometry table onto the stack

    lua_pushstring(L, "type");
    lua_gettable(L, -2);
    const char* geoType = lua_tostring(L, -1); //get the type of geometry
    lua_pop(L, 1); //pop the type from the stack

    std::cout << "the type is " << geoType << std::endl;
    if (std::strcmp(geoType, "sphere") == 0) // we constructing a sphere
    {
        std::cout << "constructing a sphere" << std::endl;
        Vector3d position = loadVector("position");
        double radius;
        loadDouble("radius", radius);
        geo = new Sphere(position, radius);
    }

    else if (std::strcmp(geoType, "plane") == 0)
    {
        std::cout << "constructing a plane" << std::endl;
        Vector3d position = loadVector("position");
        Vector3d normal = loadVector("normal");
        geo = new Plane(position, normal);
    }

    lua_pop(L, 1); //pop the geometry table
    return geo;
}
//------------------------loadLight------------------------
// load the color "color" and the vector "position" and return
// a pointer to the new light
//---------------------------------------------------------
Light* Scene::loadLight()
{
    Color3d color = loadColor("color");
    Vector3d position = loadVector("position");
    Light* light = new Light(color, position);
    return light;
}

Color3d Scene::loadColor(const char* colorName)
{
    double red=0, green=0, blue=0;
    //std::cout<<"loading color";
    lua_pushstring(L, colorName); //push the colorName string on the stack
    lua_gettable(L, -2); //and now get the color's table on the stack
    loadDouble("r", red);
    loadDouble("g", green);
    loadDouble("b", blue);
    lua_pop(L, 1); //pop the color table

    return Color3d(red, green, blue);
}

Color3d Scene::loadVector(const char* vectorName)
{
    double x, y, z;
    lua_pushstring(L, vectorName); //push the vectorName string on the stack
    lua_gettable(L, -2); //and now get the vectorName table on the stack
    loadDouble("x", x);
    loadDouble("y", y);
    loadDouble("z", z);
    lua_pop(L, 1); //pop the vector table

    return Vector3d(x, y, z);
}

//------------------------loadDouble-----------------------
// load a double from the current table.  if doubleName doesn't
// exist then return 0 and set d to 0, else, return 1
//---------------------------------------------------------
int Scene::loadDouble(const char* doubleName, double& d)
{
    std::cout<<"loading double "<<doubleName<<std::endl;
    
    lua_pushstring(L, doubleName); //push key
    lua_gettable(L, -2); //pop key and push value
    if (lua_isnil(L, -1)) { //if doubleName doesn't exist
        d = 0.0; //set default
        return 0;
  }
  d = (double) lua_tonumber(L, -1); //set the double
  lua_pop(L, 1); //remove the double
  return 1;
}