#include "BasicShader.h"


Color3d BasicShader::getColor(Ray& ray, int level) const
{
    Color3d color = Color3d::Zero();
    Body* body;
    Vector3d normal = Vector3d::Zero();
    Vector3d interPoint = Vector3d::Zero();

    int hit = scene->getBodyCollection()->intersectRay(ray, interPoint, normal, body);
    if (hit != 0)
    {
        color =  getPhong(ray, interPoint, normal, body); //phong shading color
        color += getReflected(ray, interPoint, normal, body, level); //add reflected color
        color += getRefracted(ray, interPoint, normal, body, hit, level); //add refracted color
    }
    return color;
}

//------------------------getPhong-----------------------
// We calculate the phong shading by looping through each of the
// lights and summing up the phong shadinng for each one
//---------------------------------------------------------
Color3d BasicShader::getPhong(Ray& ray, Vector3d& interPoint, Vector3d& normal, Body* body) const
{
    Color3d shading = Color3d::Zero();
    
    double bodyDiffuse = body->getMaterial()->getDiffuse();
    double bodySpecular = body->getMaterial()->getSpecular();
    double bodyShininess = body->getMaterial()->getShininess();

    for (int i=0; i<scene->getLights().size(); ++i)
    {
        Light* light = scene->getLights()[i];

        //get vector from the intersection point to the light
        Vector3d interToLight = light->getPosition() - interPoint;
        
        double dot = normal.dot(interToLight);
        //we check the dot product before normalizing so we dont waste a sqrt
        if (dot < 0) continue; //no light hits this point so go to the next light

        double interToLightSq = interToLight.squaredNorm();
        interToLight.normalize();
        dot = normal.dot(interToLight); //recalculate dot product
        
        //if we're in a shadow then stop
        if (inShadow(interPoint, interToLight, interToLightSq)) continue;

        //get vector from intersection point to viewer
        Vector3d interToViewer = - ray.getDirection();
                
        //get the direction of light reflected from the surface
        Vector3d reflectedLight = (-interToLight) + 2 * normal.dot(interToLight) * normal;
        
        Color3d colorCombo = body->getMaterial()->getColor().cwiseProduct(light->getColor());

        //if the body emits diffuse light
        if (bodyDiffuse > 0)
        {
            shading += dot * bodyDiffuse * colorCombo;           
        }
        //if the body emits specular light
        if (bodySpecular > 0)
        {
            double specDot = reflectedLight.dot(interToViewer);
            if (specDot > 0)
            {
                shading += bodySpecular * pow(specDot, bodyShininess) * light->getColor();
            }
        }
    }

    return shading;
}

//----------------------inShadow---------------------------
// Checks if the segment connecting point and the light source
// intersects a body
//---------------------------------------------------------
bool BasicShader::inShadow(Vector3d& point, Vector3d& lightDir, double distSq) const
{
    Ray shadowRay(point + REFLECT_JUMP * lightDir, lightDir);
    return scene->getBodyCollection()->hasIntersection(shadowRay, distSq);
}

//-----------------------getReflected----------------------
// Recursivly calculate the color of light that may be reflected
// in the direction of the ray
//---------------------------------------------------------
Color3d BasicShader::getReflected(Ray& ray, Vector3d& interPoint, Vector3d& normal, Body* body, int level) const
{
    Color3d color = Color3d::Zero();
    double bodyReflect = body->getMaterial()->getReflection(); //the body's reflection constant

    if ((bodyReflect < 0.0) || (level >= MAX_RECURSE)) return color;

    Vector3d reflectDir = ray.getDirection() - 2 * normal.dot(ray.getDirection()) * normal;
    Ray reflectRay(interPoint + REFLECT_JUMP*reflectDir, reflectDir);
    color = getColor(reflectRay, level+1); //recursively find the color of the reflected ray
    return bodyReflect * body->getMaterial()->getColor().cwiseProduct(color);
}

//-----------------------getRefracted----------------------
// Recursivly calculate the color of light that may be refracted
// in the direction of the ray.  If dir is 1 then the ray comes from outside
// the body. If it's -1 the it comes from inside.
//The calculations are based on the paper:
// http://www.devmaster.net/articles/raytracing_series/Reflections%20and%20Refractions%20in%20Raytracing.pdf
//---------------------------------------------------------
Color3d BasicShader::getRefracted(Ray& ray, Vector3d& interPoint, Vector3d& normal, 
                                    Body* body, int dir, int level) const
{
    Color3d color = Color3d::Zero();
    double bodyRefraction = body->getMaterial()->getRefraction(); //the body's reflection constant

    if ((bodyRefraction < 0.0) || (level >= MAX_RECURSE)) return color;

    double n1, n2; //n1 is the index of refraction of the material on the side that the normal is pointing at
    Vector3d refractNorm;
    if (dir == 1) //the ray is hitting from outside the body
    {
        n1 = 1; //the ray is coming from air
        n2 = body->getMaterial()->getRefractIndex();
        refractNorm = normal; //the normal is ok
    } else { //the ray is coming from inside the body
        n1 = body->getMaterial()->getRefractIndex();
        n2 = 1;
        refractNorm = -1 * normal; //reverse the normal
    }

    double n = n1 / n2;
    double cosI = refractNorm.dot(ray.getDirection()); //cos(angle_of_incidence)
    double sinT2 = (n*n) * (1 - cosI*cosI); //(sin(angle_of_refraction))^2

    if (sinT2 > 1.0) return color; //total internal reflection

    Vector3d refractDir = n * ray.getDirection() - (n*cosI + sqrt(1-sinT2)) * refractNorm;
    Ray refractRay(interPoint + REFLECT_JUMP*refractDir, refractDir);
    color = getColor(refractRay, level+1); //recursively find the color of the reflected ray
    return bodyRefraction * body->getMaterial()->getColor().cwiseProduct(color);
}